import click

from infra_buddy.commandline import cli
from infra_buddy.context.deploy_ctx import DeployContext
from infra_buddy.context import service_definition
from infra_buddy.context.service_definition import ServiceDefinition
from infra_buddy.deploy.cloudformation_deploy import CloudFormationDeploy
from infra_buddy.template.template import NamedLocalTemplate
from infra_buddy.utility import print_utility


@cli.command(name='generate-service-definition',
             short_help="Generate a service definition for use by the deploy-service command.")
@click.option("--service-template-directory", type=click.Path(exists=True), help="The directory containing "
                                                                         "the service template.")
@click.option("--service-type", help="The service-type that corresponds with the provided template directory (or a "
                                     "built-in service type).")
@click.pass_obj
def deploy_cloudformation(deploy_ctx, service_template_directory, service_type):
    # type: (DeployContext,str,str) -> None
    do_command(deploy_ctx, service_template_directory, service_type)


def do_command(deploy_ctx, service_template_directory=None, service_type=None, destination=None):
    # type: (DeployContext,[str or None],str) -> str
    if service_template_directory is None:
        print_utility.warn(
            "Service template directory was not provided.  Assuming service-type '{}' is built-in.".format(
                service_type))
        template = deploy_ctx.template_manager.get_known_template(template_name=service_type)
        deploy = CloudFormationDeploy(stack_name=deploy_ctx.stack_name, template=template, deploy_ctx=deploy_ctx)
    else:
        deploy = CloudFormationDeploy(stack_name=deploy_ctx.stack_name,
                                      template=NamedLocalTemplate(service_template_directory),
                                      deploy_ctx=deploy_ctx)
    return ServiceDefinition.save_to_file(application=deploy_ctx.application,
                                          role=deploy_ctx.role,
                                          deploy_params=deploy.get_default_params(),
                                          known_service_modifications=deploy_ctx.template_manager.get_service_modifications_for_service(service_type)
                                          , service_type=service_type,
                                          destination=destination)



import json
import os

from jsonschema import validate

from infra_buddy.deploy.datadog_monitor_deploy import DataDogMonitorDeploy
from infra_buddy.deploy.ecs_deploy import ECSDeploy
from infra_buddy.deploy.s3_deploy import S3Deploy
from infra_buddy.utility import print_utility

_DATADOG_PROVIDER = "datadog"

_MONITORS = "monitors"

_PROVIDER = "provider"

_ARTIFACT_FILE = "monitor.json"


class MonitorDefinition(object):
    schema = {
        "type": "object",
        "properties": {
            _PROVIDER: {"type": "string", "enum": [_DATADOG_PROVIDER]},
            _MONITORS: {
                "type": "array",
                "items":
                    {
                        "type": "object",
                        "properties": {
                            "type": {"type": "string"},
                            "query": {"type": "string"},
                            "name": {"type": "string"},
                            "message": {"type": "string"},
                            "tags": {
                                "type": "array",
                                "items": {"type": "string"}
                            },
                            "options": {
                                "type": "object",
                                "properties": {
                                    "timeout_h": {"type": "number"},
                                    "notify_no_data": {"type": "boolean"},
                                    "notify_no_data_timeframe": {"type": ["string", "null"]},
                                    "notify_audit": {"type": "boolean"},
                                    "require_full_window": {"type": "boolean"},
                                    "new_host_delay": {"type": "number"},
                                    "include_tags": {"type": "boolean"},
                                    "escalation_message": {"type": "string"},
                                    "locked": {"type": "boolean"},
                                    "renotify_interval": {"type": "number"},
                                    "evaluation_delay": {"type": "number"},
                                    "thresholds": {
                                        "type": "object",
                                        "properties": {
                                            "ok": {"type": "number"},
                                            "warning": {"type": "number"},
                                            "critical": {"type": "number"}
                                        }
                                    },
                                    "silenced": {
                                        "type": "object",
                                        "patternProperties": {
                                            "^.*$": {"type": "integer"}
                                        }
                                    }

                                }

                            }
                        },
                        "required": [
                            "type",
                            "query",
                            "name"
                        ]
                    }
            }
        }
    }

    @classmethod
    def create(cls, provider=None, monitors=None):
        # type: (str, str, str) -> MonitorDefinition
        if provider == _DATADOG_PROVIDER:
            return DatadogMonitorDefinition(monitors=monitors)
        else:
            return NOOPMonitorDefinition(None)

    @classmethod
    def create_from_directory(cls, artifact_directory):
        # type: (str) -> MonitorDefinition
        definition = MonitorDefinition._load_monitor_definition(artifact_directory)
        if definition:
            validate(definition, MonitorDefinition.schema)
            return cls.create(definition[_PROVIDER], definition[_MONITORS])
        else:
            return cls.create()

    def __init__(self, monitors):
        super(MonitorDefinition, self).__init__()
        self.monitors = monitors

    @staticmethod
    def _load_monitor_definition(artifact_directory):
        artifact_def_path = os.path.join(artifact_directory, _ARTIFACT_FILE)
        if os.path.exists(artifact_def_path):
            print_utility.info("Defining artifact definition with monitor.json - {}".format(artifact_def_path))
            with open(artifact_def_path, 'r') as art_def:
                return json.load(art_def)
        else:
            print_utility.warn("Monitor definition (monitor.json) did not exist in artifact directory."
                               " Continuing infrastructure update without monitor deploy.")
        return None

    def generate_execution_plan(self, deploy_ctx):
        raise Exception("Abstract Method not implemented")

    def save_to_file(self, destination_dir=None):
        if destination_dir:
            path = os.path.join(destination_dir, _ARTIFACT_FILE)
        else:
            path = _ARTIFACT_FILE
        print_utility.info("Persisting monitor definition - {}".format(path))
        with open(path, 'w') as file:
            json.dump(self.monitors, file)
        return path

    def register_env_variables(self, deploy_ctx):
        pass


class DatadogMonitorDefinition(MonitorDefinition):
    def generate_execution_plan(self, deploy_ctx):
        return [DataDogMonitorDeploy(self.monitors, deploy_ctx)]


class NOOPMonitorDefinition(MonitorDefinition):
    def generate_execution_plan(self, deploy_ctx):
        return None



import os
import tempfile

from infra_buddy.aws import s3 as s3util
from infra_buddy.aws.cloudformation import CloudFormationBuddy
from infra_buddy.aws.s3 import S3Buddy
from infra_buddy.deploy.deploy import Deploy
from infra_buddy.utility import print_utility
import datadog as dd


class DataDogMonitorDeploy(Deploy):
    def __init__(self, monitors, ctx):
        super(DataDogMonitorDeploy, self).__init__(ctx)
        self.monitors = monitors

    def _internal_deploy(self, dry_run):
        to_deploy = self.expand_monitors()
        for monitor in to_deploy:
            print_utility.info("Deploying datadog monitor: {}".format(monitor['name']))
            if not dry_run:
                self.init_dd()
                existing_id = self.find_monitor_if_exists(monitor['name'])
                if not existing_id:
                    response = dd.api.Monitor.create(**monitor)
                    created_name = response.get('name', None)
                    if created_name:
                        print_utility.info("Created monitor - {}".format(created_name))
                    else:
                        print_utility.error("Error creating monitor - {}".format(response), raise_exception=True)
                else:
                    response = dd.api.Monitor.update(id=existing_id, **monitor)
                    print_utility.info("Updated monitor - {}".format(response['name']))

    def init_dd(self):
        api_key = self.deploy_ctx.get('DATADOG_KEY', None)
        app_key = self.deploy_ctx.get('DATADOG_APP_KEY', None)
        if api_key is None or app_key is None:
            print_utility.error("Can not deploy datadog monitor without configuring DATADOG_KEY and DATADOG_APP_KEY",
                                raise_exception=True)
        dd.initialize(api_key=api_key, app_key=app_key)

    def expand_monitors(self):
        to_deploy = []
        for monitor in self.monitors:
            monitor = self.deploy_ctx.recursive_expand_vars(monitor)
            self.perform_data_checks(monitor)
            to_deploy.append(monitor)
        return to_deploy

    def __str__(self):
        return "{} - {} monitors".format(self.__class__.__name__, len(self.monitors))

    def perform_data_checks(self, monitor):
        # datadog exports more complex metric alerts are query alerts, but it wants them
        # created as metric alerts so help the user out if they just copy/pasted the datadog output
        if 'query alert' == monitor['type']:
            monitor['type'] = 'metric alert'
        monitor['name'] = self.deploy_ctx.expandvars("{}: {}".format("${ENVIRONMENT}-${APPLICATION}-${ROLE}", monitor['name']))

    def find_monitor_if_exists(self, name_):
        for mon in self.get_all_monitors_by_name(name_):
            if mon['name'] == name_:
                return mon['id']
        return None

    def get_all_monitors_by_name(self, name_):
        return dd.api.Monitor.get_all(name=name_)

    def delete_all_by_name(self, name):
        to_delete = self.get_all_monitors_by_name(name_=name)
        for mon in to_delete:
            dd.api.Monitor.delete(mon['id'])



from datetime import timedelta

import click

VERBOSE = False


def configure(verbose):
    global VERBOSE
    VERBOSE = verbose


def info(msg):
    if VERBOSE: click.secho(msg, fg='green')


def warn(msg):
    click.secho(msg, fg="yellow")


def progress(msg):
    click.secho(msg, fg='blue')


def banner_warn(msg, data):
    banner(msg)
    info_banner(data)


def error(err_msg, raise_exception=False):
    click.secho(err_msg, fg="red")
    if raise_exception:
        raise click.UsageError(err_msg)


def banner(data):
    click.secho(data, fg="yellow")

def info_banner(data):
    click.secho(data, fg='green')

def print_time_delta(tdelta):
    # type: (timedelta) -> object
    d = {"days": tdelta.days}
    d["hours"], rem = divmod(tdelta.seconds, 3600)
    d["minutes"], d["seconds"] = divmod(rem, 60)
    format_ = "{minutes}m {seconds}s"
    if d['days'] > 0:
        format_ = "{days} days " + format_
    if d['hours'] > 0:
        format_ = "{hours}h " + format_
    return format_.format(**d)



from infra_buddy.aws.ecs import ECSBuddy
from infra_buddy.deploy.deploy import Deploy
from infra_buddy.utility import print_utility


class ECSDeploy(Deploy):
    def __init__(self, artifact_id, artifact_location, deploy_ctx):
        super(ECSDeploy, self).__init__(deploy_ctx)
        self.artifact_id = artifact_id
        self.artifact_location = artifact_location
        self.ecs_buddy = ECSBuddy(self.deploy_ctx)

    def _internal_deploy(self, dry_run):
        self.ecs_buddy.set_container_image(self.artifact_location, self.artifact_id)
        if dry_run:
            print_utility.warn("ECS Deploy would update service {} to use image {}".format(self.ecs_buddy.ecs_service,
                                                                                           self.ecs_buddy.new_image))
            return None
        if self.ecs_buddy.requires_update():
            self.ecs_buddy.perform_update()

            print_utility.progress("ECS Deploy updated service {} to use image {}".format(self.ecs_buddy.ecs_service,
                                                                                           self.ecs_buddy.new_image))
            return True
        else:
            print_utility.progress("ECS already using image - "
                               "{artifact_location}:{artifact_id}".
                               format(artifact_location=self.artifact_location,
                                      artifact_id=self.artifact_id))
            return False

    def __str__(self):
       return "{} - {}:{}".format(self.__class__.__name__,self.artifact_location,self.artifact_id)


